#ifndef SERVO_H
#define SERVO_H

#include <alternate-gpio.hpp>
#include <gpio-port.hpp>

#include "pwm-timer.hpp"
#include "servo-init.hpp"

using namespace stm32f072xb;

// usage sample

/*!
 * @brief Construct a new Servo object
 * @code {.cpp}
 * servo(servoInit, 90, -90, 90, 450, 2300); // For the HS-422
 * @endcode
 *
 * @param servoInit Servo initialization structure
 * @param amplitude Amplitude of the servo motor
 * @param minAngle Minimum angle the servo can go to (software barrier)
 * @param maxAngle Maximum angle the servo can go to (software barrier)
 * @param pulseWidthMIN Pulse Width minimum value (differs based on the servo model, default = 900)
 * @param pulseWidthMAX Pulse Width maximum value (differs based on the servo model, default = 2100)
 */
/**
 * We found the maximum and minimum frequency range in microseconds for the following servos.
 * This is the template for displaying the informations: Servo Name : minimum pulse width-maximum pulse width
 * HS-422 : 450-2300
 * HS-485HB : 900-2100
 * MG-996R : 544-2400
 * HS-785HB :1050-1950
 * MG90S : 500-3000
 **/
template <Port pwmPort>
class Servo {
   private:
	static const uint16_t PWM_RESOLUTION = 30000;
	static const uint16_t PWM_FREQUENCY = 50;
	uint16_t _zeroDegreeDutyCycle;
	int16_t _currentAngle;
	int16_t _minAngle;
	int16_t _maxAngle;
	float _scaler;

	PwmTimer _timer;
	AlternateGpio<pwmPort> _pwmOutputPin;

   public:
	/**
	 * @brief Construct a new Servo object
	 *
	 * @param servoInit Servo initialization structure
	 * @param amplitude Amplitude of the servo motor
	 * @param minAngle Minimum angle the servo can go to (software barrier)
	 * @param maxAngle Maximum angle the servo can go to (software barrier)
	 * @param startAngle Start angle of the servo
	 * @param pulseWidthMIN Pulse Width minimum value (differs based on the servo model, default = 900)
	 * @param pulseWidthMAX Pulse Width maximum value (differs based on the servo model, default = 2100)
	 */
	Servo(ServoInit& servoInit,
	      uint16_t amplitude,
	      int16_t minAngle,
	      int16_t maxAngle,
	      uint16_t pulseWidthMIN = 900,
	      uint16_t pulseWidthMAX = 2100)
	    : _zeroDegreeDutyCycle((pulseWidthMIN + pulseWidthMAX) / 2),
	      _currentAngle((maxAngle + minAngle) / 2),
	      _minAngle(minAngle),
	      _maxAngle(maxAngle),
	      _scaler((float)(_zeroDegreeDutyCycle - pulseWidthMIN) / (float)amplitude),
	      _timer(servoInit.timer, servoInit.OCC, PWM_RESOLUTION, PWM_FREQUENCY, _zeroDegreeDutyCycle),
	      _pwmOutputPin({servoInit.pwmPin, Pupdr::NONE, servoInit.alternateFunction}) {
		setAngle((maxAngle + minAngle) / 2);
	}

	int16_t getAngle() const { return _currentAngle; }

	void setAngle(int16_t angle) {
		if (angle > _maxAngle) {
			angle = _maxAngle;
		}
		if (angle < _minAngle) {
			angle = _minAngle;
		}
		_currentAngle = angle;
		uint16_t dutyCycle = _zeroDegreeDutyCycle + _scaler * angle;
		_timer.setDutyCycle(dutyCycle);
	}

	void incrementAngle() {
		if (_currentAngle < _maxAngle) {
			setAngle(++_currentAngle);
		}
	}
	void decrementAngle() {
		if (_currentAngle > _minAngle) {
			setAngle(--_currentAngle);
		}
	}
};

#endif
